from .FRNNI import FRNNI
from .CDI import CDI
from .HotDeckImputation import HotDeckImputation
from .KNNImputer import KNNImputer
from .PMM import PMM
from .SICE import SICE
from .MICE import MICE
from .errors import *



'''
Created with love by Sigmoid

@Author - Stojoc Vladimir - vladimir.stojoc@iis.utm.md
'''
# Importing all needed libraries
import numpy as np
import pandas as pd
from sklearn.base import BaseEstimator, TransformerMixin

from .errors import NonNumericDataError


class KNNImputer(BaseEstimator, TransformerMixin):
    def __init__(self, n_neighbors = 5, epsilon = 1e-10) -> None:
        '''
            The constructor of the KNNImputer class.
        :param n_neighbors: int, default = 5
            The number of neighbors to find for every sample with missing values.
        :param metric: str, default = 'minkowski'
            The distance metric to use for finding the nearest neighbors.
        :param epsilon: int, default = 1e-10
            The parameter do add to inverse distance weight in case the distance is 0 to nearest neighbors
        '''
        self.n_neighbors = n_neighbors
        self.epsilon = epsilon


    def fit(self, X: 'np.array', y: 'np.array' = None, **fit_params: dict):
        '''
            The fit function of the KNNImputer, fits up the model.
        :param X: 2-d numpy array or pd.DataFrame
            The 2-d numpy array or pd.DataFrame that represents the feature matrix.
        :param y: 1-d numpy array or pd.DataFrame
            The 1-d numpy array or pd.DataFrame that represents the target array.
        :param fit_params: dict
            The fit parameters of the function, they are ignored in the body of the function.
        :return: KNNImputer
            The fitter KNNImputer object.
        '''
        return self

    def transform(self, X: 'np.array', y: 'np.array' = None, **fit_params: dict) -> 'np.array':
        '''
            The transform function of the KNNImputer, transforms the passed data..
        :param X: 2-d numpy array
            The 2d matrix representing the data that should be filled.
        :param y: 1-d numpy array, default = None
            The 1d array representing the target vector, it is not used in this function, ignored.
        :param fit_params: dict
            The fit parameters that control the fitting process.
        :return: np.array
            The transformed data.
        '''

        if not np.issubdtype(X.dtype, np.number):
            raise NonNumericDataError('Not all passed columns of the dataframe represent numeric data.')

        # Creating a copy of array
        new_X = X.copy()

        # Looping through all rows
        for x_idx, x in enumerate(X):
            check_nan = np.isnan(x)
            # Check for Nans in rows
            if check_nan.any():
                nan_indexes = np.where(check_nan == True)[0]
                # Go through columns with NaNs
                for i in nan_indexes:
                    # select all rows from X where the ith column in not nan
                    nan_col = X[:, i]
                    X_not_nan_flags = np.isnan(nan_col)
                    # Get all rows where this column is not NaN
                    X_not_nan_indexes = np.where(X_not_nan_flags == False)[0]
                    X_not_nan = X[X_not_nan_indexes, :]

                    # Get closest k neighbours
                    nearest_neighbors, distances = self.__get_k_neighbours(x, X_not_nan)
                    # Select the needed column from nearest neighbors
                    nearest_neighbors = nearest_neighbors[:, i]
                    # Calculating a weighted average
                    weighted_average = np.average(nearest_neighbors, weights=1/distances)

                    # Imputing calculated value
                    new_X[x_idx, i] = weighted_average

        return new_X


    def fit_transform(self, X: 'np.array', y: 'np.array' = None, **fit_params: dict):
        '''
            The transform function of the KNNImputer, transforms the passed data..
        :param X: 2-d numpy array
            The 2d matrix representing the data that should be filled.
        :param y: 1-d numpy array, default = None
            The 1d array representing the target vector, it is not used in this function, ignored.
        :param fit_params: dict
            The fit parameters that control the fitting process.
        :return: np.array
            The transformed data.
        '''
        return self.fit(X).transform(X)

    def apply(self, df: 'pd.DataFrame', columns: list = None) -> 'pd.DataFrame':
        '''
            This function allows applying the transformer on certain columns of a data frame.
        :param df: pandas DataFrame
            The data frame with the possible NaN-values that should be inputed.
        :param columns: list
            The column that should be taken into account during the NaN value imputation process.
        :return: pandas DataFrame
            The new pandas DataFrame with transformed columns.
        '''

        cols_not_in_df = [col for col in columns if col not in df]
        if cols_not_in_df:
            raise ValueError(f"{', '.join(cols_not_in_df)} are not present in the DataFrame")

        X = df[columns].values

        if not np.issubdtype(X.dtype, np.number):
            raise NonNumericDataError('Not all passed columns of the dataframe represent numeric data.')
        else:
            df_new = self.fit_transform(X)
            df.update(pd.DataFrame(df_new, columns=columns))


    def __get_k_neighbours(self, x, X_not_nan):
        '''
            KNN, getting nearest neighbors
        :param x: Numpy array
            the row from df to get neighbours from
        :param X_not_nan: Numpy.ndarray
            Samples from where we find neighbours
        '''

        # Calculating distance to all other samples
        dist = np.sqrt(np.nansum((X_not_nan - x) ** 2, axis=1))
        dist += self.epsilon
        dist_rows = list(enumerate(dist))
        # Sorting distances and first k neighbors
        dist_rows.sort(key=lambda x: x[1])
        dist_rows = dist_rows[:self.n_neighbors]

        return X_not_nan[np.array([x[0] for x in dist_rows])], np.array([x[1] for x in dist_rows])


from distutils.core import setup
long_description = '''
Reparo is a python sci-kit learn inspired package for Missing Value Imputation. It contains a some feature transformers to eliminate Missing Values (NaNs) from your data for Machine Learning Algorithms.\n
This version of reparo has the next methods of missing value imputation:\n
1) Cold-Deck Imputation (CDI).\n
2) Hot-Deck Imputation (HotDeckImputation).\n
3) Fuzzy-Rough Nearest Neighbor for Imputation (FRNNI).\n
4) K-Nearest Neighbors Imputation (KNNI).\n
5) Single Center Imputation from Multiple Chained Equation (SICE).\n
6) Predictive Mean Matching (PMM).\n
7) Multivariate Imputation by Chained Equation (MICE).\n
All these methods work like normal sklearn transformers. They have fit, transform and fit_transform functions implemented.\n
Additionally every reparo transformer has an apply function which allows to apply an transformation on a pandas Data Frame.\n
How to use reparo\n
To use a transformer from reparo you should just import the transformer from reparo in the following framework:\n
```from reparo import <class name>```\n
class names are written above in parantheses.\n
Next create a object of this algorithm (I will use k-Nearest Neighbors Imputation as an example).\n
```method = KNNI()```\n
Firstly you should fit the transformer, passing to it a feature matrix (X) and the target array (y). y argument is not really used (as it causes data leackage)\n
```method.fit(X, y)```\n
After you fit the model, you can use it for transforming new data, using the transform function. To transform function you should pass only the feature matrix (X).\n
```X_transformed = method.transform(X)```\n
Also you can fit and transform the data at the same time using the fit_transform function.\n
```X_transformed = method.fit_transform(X)```\n
Also you can apply a transformation directly on a pandas DataFrame, choosing the columns that you want to change.\n
```new_df = method.apply(df, 'target', ['col1', 'col2'])```\n
With love from Sigmoid.\n
We are open for feedback. Please send your impression to papaluta.vasile@isa.utm.md\n
'''
setup(
  name = 'reparo',
  packages = ['reparo'],
  version = '0.1.0',
  license='MIT',
  description = 'Reparo is a python sci-kit learn inspired package for Missing Value Imputation.',
  long_description=long_description,
  author = 'SigmoidAI - Păpăluță Vasile',
  author_email = 'vpapaluta06@gmail.com',
  url = 'https://github.com/SigmoidAI/reparo',
  download_url = 'https://github.com/ScienceKot/kydavra/archive/v1.0.tar.gz',    # I explain this later on
  keywords = ['ml', 'machine learning', 'feature engineering', 'python', 'data science'],   # Keywords that define your package best
  install_requires=[            # I get to this in a second
          'numpy',
          'pandas',
          'scikit-learn',
          'statsmodels'
      ],
  classifiers=[
    'Development Status :: 3 - Alpha',
    'Framework :: Jupyter',
    'Intended Audience :: Science/Research',
    'Topic :: Scientific/Engineering :: Artificial Intelligence',
    'Topic :: Scientific/Engineering :: Mathematics',
    'Topic :: Software Development :: Libraries :: Python Modules',
    'Topic :: Software Development :: Build Tools',
    'License :: OSI Approved :: MIT License',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.4',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: 3.7',
    'Programming Language :: Python :: 3.8',
    'Programming Language :: Python :: 3.9',
    'Programming Language :: Python :: 3.10',
    'Programming Language :: Python :: 3.11',
  ],
)
